import arlunio as ar
import arlunio.mask as mask
import arlunio.math as math


@ar.definition
def Checker(x: math.X, y: math.Y) -> mask.Mask:
    """
    .. arlunio-image:: Checker
       :align: center

       ::

          from arlunio.pattern import Checker
          from arlunio.image import fill

          checker = Checker()
          image = fill(checker(width=256, height=256))

    A simple checker pattern
    """
    return x * y > 0



import functools
import logging
from typing import Union

import numpy as np

import arlunio as ar


class Mask(np.ndarray):
    """A mask is just a boolean numpy array.

    They are typically used to represent 'selections' for various operations such as
    when coloring a region of an image.
    """

    def __new__(cls, arr):
        return np.asarray(arr).view(cls)

    def __mul__(self, other):

        try:
            return np.logical_and(self, other)
        except ValueError:

            mask = self.copy()
            size = np.prod(other.shape)

            if mask[mask].shape == size:
                mask[mask] = other.reshape(size)
                return mask

            raise

    def __neg__(self):
        return np.logical_not(self)

    def __sub__(self, other):
        return np.logical_and(self, np.logical_not(other))

    def __rsub__(self, other):
        return np.logical_and(other, np.logical_not(self))

    @classmethod
    def empty(cls, *shape):
        """Return an empty mask with the given shape.

        Example
        -------
        >>> from arlunio.mask import Mask
        >>> Mask.empty(3, 4)
        Mask([[False, False, False, False],
              [False, False, False, False],
              [False, False, False, False]])
        """

        if len(shape) == 1 and isinstance(shape[0], tuple):
            return cls(np.full(shape[0], False))

        return cls(np.full(shape, False))

    @classmethod
    def full(cls, *shape):
        """Return a full mask with the given shape.

        Example
        -------
        >>> from arlunio.mask import Mask
        >>> Mask.full(3, 4)
        Mask([[ True,  True,  True,  True],
              [ True,  True,  True,  True],
              [ True,  True,  True,  True]])
        """

        if len(shape) == 1 and isinstance(shape[0], tuple):
            return cls(np.full(shape[0], True))

        return cls(np.full(shape, True))


@ar.definition
def Empty(width: int, height: int) -> Mask:
    """An empty mask.

    Example
    -------
    >>> from arlunio.mask import Empty
    >>> empty = Empty()
    >>> empty(width=4, height=3)
    Mask([[False, False, False, False],
          [False, False, False, False],
          [False, False, False, False]])
    """
    return Mask.empty(height, width)


@ar.definition
def Full(width: int, height: int) -> Mask:
    """A full mask.

    Example
    -------
    >>> from arlunio.mask import Full
    >>> full = Full()
    >>> full(width=4, height=3)
    Mask([[ True,  True,  True,  True],
          [ True,  True,  True,  True],
          [ True,  True,  True,  True]])
    """
    return Mask.full(height, width)


@ar.definition(operation=ar.Defn.OP_ADD)
def MaskAdd(
    width: int, height: int, *, a: ar.Defn[Mask] = None, b: ar.Defn[Mask] = None
) -> Mask:
    """Add any two mask producing definitions together.

    The resulting defintion will return a mask that is :code:`True` if a given point
    is :code:`True` in either :code:`a` or :code:`b`.

    Attributes
    ----------
    a:
        The first mask
    b:
        The second mask
    """

    return a(width=width, height=height) + b(width=width, height=height)


@ar.definition(operation=ar.Defn.OP_SUB)
def MaskSub(
    width: int, height: int, *, a: ar.Defn[Mask] = None, b: ar.Defn[Mask] = None
) -> Mask:
    """Subtract one mask away from another mask.

    The resulting definition will return a mask that is :code:`True` only if a given
    point is in :code:`a` **and not** in :code:`b`.

    .. note::

       You will get very different results depending on which way around you set
       :code:`a` and :code:`b`!

    Attributes
    ----------
    a:
        The first "base" mask
    b:
        The second mask that defines the region to remove from :code:`a`
    """
    return a(width=width, height=height) - b(width=width, height=height)


@ar.definition(operation=ar.Defn.OP_MUL)
def MaskMul(
    width: int, height: int, *, a: ar.Defn[Mask] = None, b: ar.Defn[Mask] = None
) -> Mask:
    """Muliply any two mask producing definitions together.

    The resulting definition will return a mask that is :code:`True` only if a given
    point is in both :code:`a` and :code:`b`.

    Attributes
    ----------
    a:
        The first mask
    b:
        The second mask
    """
    return a(width=width, height=height) * b(width=width, height=height)


def any_(*args: Union[bool, np.ndarray, Mask]) -> Mask:
    """Given a number of conditions, return :code:`True` if any of the conditions
    are true.

    This function is implemented as a thin wrapper around numpy's
    :data:`numpy:numpy.logical_or` function so that it can take an arbitrary number of
    inputs. This also means that this function will accept arrays of differing sizes,
    assuming that they can be broadcasted to a common shape.

    Parameters
    ----------
    args:
      A number of boolean conditions, a condition can either be a single boolean value
      or a numpy array of boolean values.

    Examples
    --------

    >>> import arlunio.mask as mask
    >>> mask.any_(True, False, False)
    Mask(True)
    >>> mask.any_(False, False, False, False)
    Mask(False)

    If the arguments are boolean numpy arrays, then the any condition is applied
    element-wise

    >>> import numpy as np
    >>> x1 = np.array([True, False, True])
    >>> x2 = np.array([False, False, True])
    >>> x3 = np.array([False, True, False])
    >>> mask.any_(x1, x2, x3)
    Mask([ True,  True,  True])

    The arguments can be any mixture of booleans, arrays and masks.

    >>> mask.any_(
    ...     False,
    ...     mask.Mask([True, False]),
    ...     np.array([[False, True], [True, False]])
    ... )
    Mask([[ True,  True],
          [ True, False]])


    See Also
    --------

    :doc:`numpy:user/basics.broadcasting`
       Numpy documentation on broadcasting.

    :doc:`numpy:user/theory.broadcasting`
       Further background on broadcasting.

    :data:`numpy:numpy.logical_or`
       Reference documentation on the :code:`numpy.logical_or` function
    """
    return Mask(functools.reduce(np.logical_or, args))


def all_(*args: Union[bool, np.ndarray, Mask]) -> Mask:
    """Given a number of conditions, return :code:`True` only if **all**
    of the given conditions are true.

    This function is implemented as a thin wrapper around numpy's
    :data:`numpy:numpy.logical_and` function so that it can take an arbitrary number of
    inputs. This also means that this function will accept arrays of differing sizes,
    assuming that they can be broadcasted to a common shape.

    Parameters
    ----------
    args:
      A number of boolean conditions, a conditon can either be a single boolean value,
      or a numpy array of boolean values.

    Examples
    --------

    >>> import arlunio.mask as mask
    >>> mask.all_(True, True, True)
    Mask(True)
    >>> mask.all_(True, False, True, True)
    Mask(False)

    If the arguments are boolean numpy arrays, then the any condition is applied
    element-wise

    >>> import numpy as np
    >>> x1 = np.array([True, False, True])
    >>> x2 = np.array([False, False, True])
    >>> x3 = np.array([False, True, True])
    >>> mask.all_(x1, x2, x3)
    Mask([False, False,  True])

    Arugments can be any mixture of booleans, masks and numpy arrays.

    >>> mask.all_(
    ...     True,
    ...     mask.Mask([True, False]),
    ...     np.array([[False, True], [True, False]])
    ... )
    Mask([[False, False],
          [ True, False]])


    See Also
    --------

    :doc:`numpy:user/basics.broadcasting`
       Numpy documentation on broadcasting.

    :doc:`numpy:user/theory.broadcasting`
       Further background on broadcasting.

    :data:`numpy:numpy.logical_and`
       Reference documentation on the :code:`logical_and` function.
    """
    return Mask(functools.reduce(np.logical_and, args))


@ar.definition
def Repeat(width: int, height: int, *, n=4, m=None, defn=None) -> Mask:
    """Given a mask producing definition, replicate the resulting mask in a grid.

    .. arlunio-image:: Simple Grid

       ::

          import arlunio.image as image
          import arlunio.mask as mask
          import arlunio.shape as shape

          pattern = mask.Repeat(defn=shape.Circle())
          img = image.fill(pattern(width=256, height=256))

    When evaluated this will create a mask with the given :code:`width` and
    :code:`height`. It will then subdivide it into an :math:`n \\times m` grid where
    each cell contains a copy of the mask as produced by the definition specified with
    the :code:`defn` attribute.

    It's important to note that the given definition must only take :code:`width` and
    :code:`height` as inputs.


    .. note::

       Due to a limitation in the current implementation, you will get the best results
       if your :math:`n \\times m` grid divides cleanly into the resolution of the final
       mask. Otherwise you will find that the generated grid won't completly fill it.

    Attributes
    ----------
    n:
        The number of times to repeat the given definition horizontally
    m:
        The number of times to repeat the given definition vertically. If :code:`None`
        this defaults to the value of :code:`n`
    defn:
        The instance of the definition to replicate.

    Examples
    --------

    .. arlunio-image:: Circular Pattern
       :gallery: examples
       :include-code:
       :width: 50%

       A pattern generated from circles::

          import arlunio as ar
          import numpy as np

          import arlunio.image as image
          import arlunio.math as math
          import arlunio.mask as mask
          import arlunio.shape as shape

          @ar.definition
          def Template(x: math.X, y: math.Y) -> mask.Mask:
              c = shape.Circle(xc=0.4, yc=0.4, pt=0.02)
              return c(x=np.abs(x), y=np.abs(y))

          pattern = mask.Repeat(defn=Template(scale=1.))
          img = image.fill(
              pattern(width=1080, height=1080), background="#000", foreground="#ff0"
          )

    .. arlunio-image:: Checkerboard
       :gallery: examples
       :include-code:
       :width: 50%

       A checkerboard::

          import arlunio.image as image
          import arlunio.mask as mask
          import arlunio.pattern as pattern

          grid = mask.Repeat(defn=pattern.Checker())
          img = image.fill(grid(width=1080, height=1080), background="white")

    """
    if m is None:
        m = n

    bg = np.full((height, width), False)

    # Draw the shape at a size determined by the size of the grid
    s_height, s_width = height // m, width // n
    mask = defn(width=s_width, height=s_height)

    # Let numpy handle the repeating of the shape across the image.
    pattern = np.tile(mask, (m, n))

    # Apply the pattern to the background, depending on the grid size and
    # image dimensions align, the generated grid may not perfectly fill the
    # image.
    p_height, p_width = pattern.shape

    bg[:p_height, :p_width] = pattern

    return bg


@ar.definition
def Map(width: int, height: int, *, layout=None, legend=None, fill=None) -> Mask:
    """Build a mask composed out of smaller, simpler masks.

    When evaluated this will produce a mask with the given :code:`width` and
    :code:`height` and divide it into a grid. The dimensions of this grid are determined
    by shape of the :code:`layout` array.

    The :code:`layout` attribute should be set to a 2D array the elements of which can
    be anything. While the :code:`legend` attribute is set to a dictionary whose keys
    correspond to values in the :code:`layout` array. These keys should then map to mask
    producing definitions. It's important to note that these definitions can only take
    :code:`width` and :code:`height` as inputs.

    The cells in the grid will then be set to the mask produced by the definition
    corresponding to the value in the :code:`layout`. If however the :code:`legend` does
    not contain a matching key then the :code:`fill` definition will be used instead.

    .. note::

       Due to a limitation in the current implementation, you will get best results
       if the dimensions of the :code:`layout` grid divide cleanly into the dimensions
       of the final mask.

    Attributes
    ----------
    fill:
        The definition to use in any cell where a corresponding definition cannot be
        found in the legend. If :code:`None` this will default to
        :class:`arlunio.mask.Empty`
    layout:
        A 2D numpy array of values representing keys from the :code:`legend` detailing
        which mask should be used in which cell.
    legend:
        A dictionary with keys corresponding to values in the :code:`layout` that map to
        mask producing definitions that should be used.

    Example
    -------
    .. arlunio-image:: Simple Map
       :include-code:
       :width: 50%
       :gallery: examples

       ::

          import arlunio.image as image
          import arlunio.mask as mask
          import arlunio.shape as shape

          import numpy as np

          top = shape.Rectangle(size=0.2, yc=1, ratio=50)
          left = shape.Rectangle(size=0.2, xc=-1, ratio=1/50)
          right = shape.Rectangle(size=0.2, xc=1, ratio=1/50)
          bottom = shape.Rectangle(size=0.2, yc=-1, ratio=50)

          legend = {
              "tt": top,
              "bb": bottom,
              "ll": left,
              "rr": right,
              "tl": top + left,
              "tr": top + right,
              "bl": bottom + left,
              "br": bottom + right
          }

          layout = np.array([
              ["tt", "tt", "tt", "tt", "tr"],
              [  "", "tl", "tt", "tr", "rr"],
              [  "", "ll", "bl", "br", "rr"],
              [  "", "bl", "bb", "bb", "br"],
              [  "",   "",   "",   "",   ""]
          ])

          map_ = mask.Map(legend=legend, layout=layout)
          img = image.fill(
              map_(width=1080, height=1080), foreground="blue", background="white"
          )
    """
    fill = fill if fill is not None else Empty()

    # TODO: Handle divisions with rounding errors
    nx, ny = len(layout), len(layout[0])
    size = {"height": height // ny, "width": width // nx}

    # Build a new dict with the values being the shapes drawn at the appropriate res
    # to ensure we only draw them once.
    items = {k: v(**size) for k, v in legend.items()}
    default = fill(**size)

    return np.block([[items.get(key, default) for key in row] for row in layout])


@ar.definition
def Pixelize(width: int, height: int, *, mask=None, defn=None, scale=16) -> Mask:
    """Produce a pixelated version of the given mask.

    .. arlunio-image:: Pixelise
       :align: center

       ::

          import arlunio.image as image
          import arlunio.mask as mask
          import arlunio.shape as shape

          pix = mask.Pixelize(defn=shape.Circle())
          img = image.fill(pix(width=256, height=256))

    This definition can either be given an existing :code:`mask` or a mask producing
    definition which can be given with the :code:`defn` attribute. Note that this
    definition can only take :code:`width` and :code:`height` as inputs.

    .. note::

       There is a limitation in the current implementation where the resulting mask may
       be smaller than expected due to rounding errors. For best results

       - Ensure that the shape of the mask given with the :code:`mask` attribute cleanly
         divides your desired :code:`width` and :code:`height`.

       - When using the :code:`defn` attribute ensure that the :code:`scale` attribute
         cleanly divides your desired :code:`width` and :code:`height`.


    Attributes
    ----------
    mask:
        The mask to pixelise. If given then :code:`defn` must be :code:`None`.
    defn:
        The mask producing definition to use. If given then :code:`mask` must be
        :code:`None`
    scale:
        When providing the :code:`defn` attribute this controls the resolution the
        definition is rendered at. Has no effect when providing a :code:`mask`

    Examples
    --------

    This definition can be used to render a mask at a higher resolution

     .. arlunio-image:: Creeper
       :gallery: examples
       :include-code:
       :width: 50%

       ::

          import numpy as np

          import arlunio.image as image
          import arlunio.mask as mask

          face = np.array([
              [False, False, False, False, False, False, False, False],
              [False, False, False, False, False, False, False, False],
              [False,  True,  True, False, False,  True,  True, False],
              [False,  True,  True, False, False,  True,  True, False],
              [False, False, False,  True,  True, False, False, False],
              [False, False,  True,  True,  True,  True, False, False],
              [False, False,  True,  True,  True,  True, False, False],
              [False, False,  True, False, False,  True, False, False],
          ])

          defn = mask.Pixelize(mask=face)
          img = image.fill(defn(width=512, height=512), background="white")

    We can also generate the mask directly from another definition.

    .. arlunio-image:: Ghost
       :include-code:
       :width: 50%
       :gallery: examples

       ::

          import arlunio as ar
          import arlunio.image as image
          import arlunio.mask as mask
          import arlunio.math as math
          import arlunio.shape as shape

          import numpy as np

          @ar.definition
          def Ghost(x: math.X, y: math.Y) -> mask.Mask:
              head = shape.Circle(yc=0.5, r=0.7)
              eyes = shape.Circle(xc=0.2, yc=0.6, r=0.3)

              body = mask.all_(
                  y < 0.5,
                  np.abs(x) < 0.49,
                  0.1 * np.cos(5 * np.pi * x) - 0.3 < y
              )

              return (head(x=x, y=y) - eyes(x=np.abs(x), y=y)) + body

          ghost = mask.Pixelize(defn=Ghost(y0=-0.3), scale=32)
          img = image.fill(
              ghost(width=1080, height=1080), foreground="#f00", background="white"
          )

    """
    logger = logging.getLogger(__name__)

    if defn is None and mask is None:
        raise ValueError("You must provide a mask or a mask producing definition.")

    if mask is not None:
        w, h = len(mask), len(mask[0])

    if defn is not None:
        # Based on the given resolution, calculate the size of each enlarged element.
        ratio = width / height

        w, h = int(scale * ratio), scale
        mask = defn(width=w, height=h)

    n, m = int(width // w), int(height // h)

    logger.debug("Mask size: (%s, %s)", w, h)
    logger.debug("Pixel size: (%s, %s)", n, m)

    fill = Mask.full(m, n)
    empty = Mask.empty(m, n)

    return Mask(np.block([[fill if col else empty for col in row] for row in mask]))



import arlunio.image as image
import arlunio.shape as shape

w, h = 1920, 1080
img = image.new(w, h, color="darkorange")

sun = shape.Circle(xc=-1.2, yc=0, r=0.6)
img += image.fill(sun(width=w, height=h), foreground="yellow")

hill = shape.Circle(xc=-1, yc=-2.1, r=1.5)
img += image.fill(hill(width=w, height=h), foreground="forestgreen")

hill = shape.Circle(xc=1, yc=-1.8, r=1.3)
img += image.fill(hill(width=w, height=h), foreground="green")


import os
import pathlib
import re
import shutil
import textwrap
from typing import Iterable
from typing import List
from typing import Set
from typing import Union

import docutils.nodes as nodes
import docutils.writers as writers
import nbformat.v4 as nbf
from docutils.io import StringOutput
from docutils.parsers import rst
from docutils.parsers.rst.directives.admonitions import BaseAdmonition
from sphinx.builders import Builder
from sphinx.util import logging
from sphinx.util.docutils import new_document

import arlunio
from .image import arlunio_image

logger = logging.getLogger(__name__)
RESOURCE_DIR = "resources"


class nbtutorial(nodes.General, nodes.Element):
    pass


def visit_nbtutorial(self, node):
    pass


def depart_nbtutorial(self, node):
    pass


class nbsolution(nodes.General, nodes.Element):
    pass


def visit_nbsolution(self, node):
    self.body.append('<details class="admonition note">\n')
    self.body.append(
        '<summary class="admonition-title">Solution (click to expand)</summary>\n'
    )


def depart_nbsolution(self, node):
    self.body.append("</details>\n")


class NBTutorialDirective(rst.Directive):
    def run(self):
        return [nbtutorial("")]


class NBSolutionDirective(BaseAdmonition):

    has_content = True
    node_class = nbsolution

    def run(self):
        (soln,) = super().run()
        return [soln]


class NotebookWriter(writers.Writer):
    """A class that walks a doctree converting it into a notebook."""

    def __init__(self, builder):
        super().__init__()
        self.builder = builder

    def translate(self):

        visitor = NotebookTranslator(self.document)
        self.document.walkabout(visitor)
        self.output = visitor.astext()


class PythonTranslator(nodes.NodeVisitor):
    """Coverts an rst doctree into a valid python file.
    Currently only used to write tutorial solutions to a python file that can be
    imported via the `%load` magic in a jupyter notebook.
    """

    def __init__(self, document):
        super().__init__(document)
        self.is_code = False
        self.source = ""

    def astext(self):
        return self.source

    # ------------------------- Visitors --------------------------------------

    def visit_Text(self, node: nodes.Text) -> None:
        source = node.astext()
        self.source += textwrap.indent(source, "# " * self.is_code)

    def depart_Text(self, node: nodes.Text) -> None:
        pass

    def unknown_visit(self, node: nodes.Node) -> None:
        pass

    def unknown_departure(self, node: nodes.Node) -> None:
        pass


class NotebookTranslator(nodes.NodeVisitor):
    """Converts rst nodes into their output representation."""

    def __init__(self, document):
        super().__init__(document)
        self.cells = []
        self.level = 0
        self.prefix = None

    def asnotebook(self):
        return nbf.new_notebook(cells=self.cells)

    @property
    def current_cell(self):

        if len(self.cells) == 0:
            return None

        return self.cells[-1]

    def new_cell(self, cell_type: str) -> None:
        current = self.current_cell

        if current is not None and current.cell_type == cell_type:
            return

        types = {"markdown": nbf.new_markdown_cell, "code": nbf.new_code_cell}
        new_cell = types[cell_type]
        self.cells.append(new_cell())

    def append(self, text):

        if self.prefix is not None:
            text = text.replace("\n", "\n" + self.prefix)

            self.current_cell.source += text
            return

        self.current_cell.source += text

    def astext(self):
        notebook = nbf.new_notebook(cells=self.cells)
        return nbf.writes(notebook)

    # --------------------------- Visitors ------------------------------------

    def _no_op(self, node):
        pass

    no_op = _no_op, _no_op

    visit_arlunio_image, depart_arlunio_image = no_op
    visit_compound, depart_compound = no_op
    visit_compact_paragraph, depart_compact_paragraph = no_op
    visit_document, depart_document = no_op
    visit_figure, depart_figure = no_op
    visit_legend, depart_legend = no_op
    visit_nbtutorial, depart_nbtutorial = no_op

    def _italics(self, node):
        self.current_cell.source += "*"

    italics = _italics, _italics

    visit_emphasis, depart_emphasis = italics
    visit_caption, depart_caption = italics

    def visit_bullet_list(self, node: nodes.bullet_list) -> None:
        self.new_cell("markdown")
        self.current_cell.source += "\n"

    def depart_bullet_list(self, node: nodes.bullet_list) -> None:
        self.current_cell.source += "\n"

    def visit_comment(self, node: nodes.comment) -> None:
        self.new_cell("markdown")

    def depart_comment(self, node: nodes.comment) -> None:
        pass

    def visit_image(self, node: nodes.image) -> None:
        self.new_cell("markdown")

        path = pathlib.Path(RESOURCE_DIR, pathlib.Path(node["uri"]).name)
        self.current_cell.source += f"\n![]({path})\n"

    def depart_image(self, node: nodes.image) -> None:
        pass

    def visit_inline(self, node: nodes.inline) -> None:
        self.current_cell.source += "["

    def depart_inline(self, node: nodes.inline) -> None:
        self.current_cell.source += "]"

    def visit_list_item(self, node: nodes.list_item) -> None:
        self.append("- ")

    def depart_list_item(self, node: nodes.list_item) -> None:
        pass

    def visit_literal(self, node: nodes.literal) -> None:
        self.current_cell.source += "`"

    def depart_literal(self, node: nodes.literal) -> None:
        self.current_cell.source += "`"

    def visit_literal_block(self, node: nodes.literal_block) -> None:
        self.new_cell("code")

    def depart_literal_block(self, node: nodes.literal_block) -> None:
        pass

    def visit_nbsolution(self, node: nbsolution) -> None:
        self.new_cell("code")

    def depart_nbsolution(self, node: nbsolution) -> None:
        pass

    def visit_note(self, node: nodes.note) -> None:
        self.new_cell("markdown")
        self.append("> **Note**\n")
        self.prefix = "> "

    def depart_note(self, node: nodes.note) -> None:
        self.prefix = None

    def visit_paragraph(self, node: nodes.paragraph) -> None:
        self.new_cell("markdown")

        if isinstance(node.parent, nodes.list_item):
            return

        self.append("\n")

    def depart_paragraph(self, node: nodes.paragraph) -> None:
        self.append("\n")

    def visit_reference(self, node: nodes.reference) -> None:
        self.append("[")

    def depart_reference(self, node: nodes.reference) -> None:
        attrs = node.attributes

        url = attrs["refuri"]
        self.append(f"]({url})")

    def visit_section(self, node: nodes.section) -> None:
        self.level += 1
        self.new_cell("markdown")

    def depart_section(self, node: nodes.section) -> None:
        self.level -= 1

    def visit_strong(self, node: nodes.strong) -> None:
        self.current_cell.source += "**"

    def depart_strong(self, node: nodes.strong) -> None:
        self.current_cell.source += "**"

    def visit_target(self, node: nodes.target) -> None:
        pass

    def depart_target(self, node: nodes.target) -> None:
        pass

    def visit_Text(self, node: nodes.Text) -> None:

        if isinstance(node.parent, nodes.comment):
            return

        if self.current_cell.cell_type == "markdown":
            self.append(node.astext())
            return

        source = node.astext()

        if ">>>" not in source:
            self.append(source)
            return

        pattern = re.compile("^(>>>|\\.\\.\\.) ?")

        def clean_line(line):
            return pattern.sub("", line)

        cleaned_source = "\n".join(
            [clean_line(line) for line in source.split("\n") if pattern.match(line)]
        )

        self.append(cleaned_source)

    def depart_Text(self, node: nodes.Text) -> None:
        pass

    def visit_title(self, node: nodes.title) -> None:
        title = "#" * self.level
        self.append(f"{title} ")

    def depart_title(self, node: nodes.title) -> None:
        self.append("\n")


def codeblock(source: str) -> nodes.literal_block:
    """Construct a valid code block."""

    block = nodes.literal_block()
    block.children = [nodes.Text(source)]

    return block


class NotebookGalleryBuilder(Builder):
    """Builder that exports gallery images from the docs as a folder of notebooks."""

    name = "nbgallery"
    format = "ipynb"

    def init(self) -> None:
        logger.debug("[nbgallery]: Outdir: %s", self.outdir)

        # Standard metadata for exported notebooks.
        self.nbmeta = {
            "author": {"name": self.app.config.author,},
            "version": arlunio.__version__,
        }

        if self.app.config.arlunio_github_author is not None:
            self.nbmeta["author"]["github"] = self.app.config.arlunio_github_author

    def get_outdated_docs(self) -> Union[str, Iterable[str]]:
        """Not too sure what should put here yet."""
        return self.env.found_docs

    def get_target_uri(self, docname: str, type: str = None) -> str:
        return docname + ".ipynb"

    def prepare_writing(self, docnames: Set[str]):
        self.docwriter = NotebookWriter(self)

    def write_doc(self, docname: str, doctree: nodes.Node) -> None:
        """Write out any gallery images on the given page."""

        logger.debug("[nbgallery]: Processing: %s", docname)
        images = [
            n for n in doctree.traverse(condition=arlunio_image) if "gallery" in n
        ]

        for image in images:

            gallery = image["gallery"]
            name = image["name"]

            code = nbf.new_code_cell(source=image["source"])
            caption = nbf.new_markdown_cell(source=image["caption"])

            metadata = {"arlunio": self.nbmeta}
            notebook = nbf.new_notebook(metadata=metadata, cells=[caption, code])

            outpath = pathlib.Path(self.outdir, gallery, name + ".ipynb")
            outpath.parent.mkdir(exist_ok=True)

            with outpath.open(mode="w") as f:
                logger.debug("[nbgallery]: Writing: %s", outpath)
                f.write(nbf.writes(notebook))


class NotebookTutorialBuilder(Builder):
    """Builder that can convert static tutorials into an interactive jupyer
    notebook."""

    name = "nbtutorial"
    format = "ipynb"

    def init(self) -> None:
        """Any initialization goes here."""
        logger.debug(f"[nbtutorial]: Outdir is: {self.outdir}")

    def get_outdated_docs(self) -> Union[str, Iterable[str]]:
        """Not too sure what we should do here yet."""

        return self.env.found_docs

    def get_target_uri(self, docname: str, type: str = None) -> str:
        """Another method to figure out."""

        uri = docname + ".ipynb"

        logger.debug(f"[nbtutorial]: Target URI: {uri}")

        return uri

    def _process_solutions(
        self, docname: pathlib.Path, solutions: List[nbsolution]
    ) -> None:
        """Given the solutions for a given tutorial save them to the resources dir.

        This also rewrites the doctree so that the solutions are replaced by cells
        the :code:`%load` magic so that the user can load the results in.
        """
        logger.debug("[nbtutorial]: Processing solutions for %s", docname)

        if len(solutions) == 0:
            return

        soln_dir = pathlib.Path(self.outdir, docname.parent, RESOURCE_DIR)
        soln_name = docname.stem

        if not soln_dir.exists():
            soln_dir.mkdir(parents=True)

        for idx, soln in enumerate(solutions):
            logger.debug("[nbtutorial]: %s", soln)

            soln_fname = f"{soln_name}-{idx + 1:02d}.py"
            soln_path = soln_dir / soln_fname

            # We need to wrap solutions inside a document for some reason
            doc = new_document("")
            doc += soln

            # Convert the solution to a valid Python file
            translator = PythonTranslator(doc)
            doc.walkabout(translator)
            python_soln = translator.astext()

            # Insert a code block into the notebook that will load the solution
            soln.children = [
                codeblock(f"%load {soln_path.relative_to(soln_dir.parent)}")
            ]

            # Write the actual solution to the given file on disk
            soln_file = os.path.join(soln_dir, soln_fname)
            logger.debug("[nbtutorial]: --> %s", soln_path)

            with open(soln_file, "w") as f:
                f.write(python_soln)

    def _process_images(self, docname: pathlib.Path, images: List[nodes.image]) -> None:
        """Given the images for a given tutorial, save them to the resources dir."""
        logger.debug("[nbtutorial]: Processing images for %s", docname)

        if len(images) == 0:
            return

        img_dir = pathlib.Path(self.outdir, docname.parent, RESOURCE_DIR)

        if not img_dir.exists():
            img_dir.mkdir(parents=True)

        for img in images:
            fname = pathlib.Path(img["uri"]).name

            source = pathlib.Path(self.app.confdir, img["uri"])
            destination = pathlib.Path(img_dir, fname)

            shutil.copy(source, destination)

    def prepare_writing(self, docnames: Set[str]) -> None:
        """A place we can add logic to?"""
        self.docwriter = NotebookWriter(self)

    def write_doc(self, docname: str, doctree: nodes.Node) -> None:
        logger.debug(f"[nbtutorial]: Called on {docname}")

        # Determine if the document represents a tutorial.
        tutorial = list(doctree.traverse(condition=nbtutorial))

        if len(tutorial) == 0:
            return

        docname = pathlib.Path(docname)

        base, fname = docname.parent, docname.stem
        logger.debug("[nbtutorial]: Base: %s, Filename: %s", base, fname)
        basedir = pathlib.Path(self.outdir, base)

        if not basedir.exists():
            basedir.mkdir(parents=True)

        # Find and write out any solutions.
        solutions = list(doctree.traverse(condition=nbsolution))
        self._process_solutions(docname, solutions)

        # Find and copy over any images
        images = list(doctree.traverse(condition=nodes.image))
        self._process_images(docname, images)

        destination = StringOutput(encoding="utf-8")
        self.docwriter.write(doctree, destination)

        pathlib.Path(basedir, "__init__.py").touch()
        outfile = os.path.join(basedir, fname + ".ipynb")

        with open(outfile, "w") as f:
            f.write(self.docwriter.output)


def register(app):

    app.add_node(
        nbtutorial,
        html=(visit_nbtutorial, depart_nbtutorial),
        latex=(visit_nbtutorial, depart_nbtutorial),
        text=(visit_nbtutorial, depart_nbtutorial),
    )

    app.add_node(nbsolution, html=(visit_nbsolution, depart_nbsolution))

    app.add_builder(NotebookGalleryBuilder)
    app.add_builder(NotebookTutorialBuilder)

    app.add_config_value("arlunio_github_author", None, "env")

    app.add_directive("nbsolution", NBSolutionDirective)
    app.add_directive("nbtutorial", NBTutorialDirective)
